!-----------------------
! madgraph - a Feynman Diagram package by Tim Stelzer and Bill Long 
! (c) 1993
!
! Filename: driver.f
!-----------------------

c Extenxion for AutoDipole by Kouhei Hasegawa, Sven Moch, and Peter Uwer, 
c Added subroutines: square_mod flowscomponents colorcomponents flowsmodcomponents
c 2009

      program MADGRAPH
!*************************************************************************
!     Main program for MadGraph
!*************************************************************************
      implicit none

! Constants

      include 'params.inc'


! Local Variables

      integer tops(0:4,0:maxnodes,0:maxtops)
      integer graphs(0:maxgraphs,0:maxlines) !graphs, topology internal lines
      integer next,i,ntopold,ii,j
      double precision sflows(maxflows,maxflows)
      logical more, lpossible
      integer fortran

c kouhei lastmod
      logical dipole,colorext,amphel
      integer emit(2,3),spec(2,3)
c kouhei lastmod

! External Functions
      logical qmatch_check

! Global Variables

      integer        iline(-maxlines:maxlines),idir(-maxlines:maxlines)
      integer        this_coup(max_coup) ,goal_coup(max_coup)
      common/to_proc/iline,idir,this_coup,goal_coup

      character*25        name
      integer        iname
      common/to_name/iname,name
      integer alen(maxgraphs)
      integer wnum(-maxlines:maxlines),wcount,wlen(maxgraphs*maxlines)
      common/toopt/wcount,wnum,wlen,alen
      integer flows(0:2*maxlines,0:maxfactors,0:24,0:maxflows)
      integer         graphcolor(0:2,0:maxflows,0:maxgraphs)
      common/to_color/graphcolor,flows
      character*(4*max_particles) particle(4)
      integer                               charge_c(max_particles)
      integer iparticle(0:max_particles,0:4),inverse(max_particles)
      common/to_model/iparticle,  particle,  inverse, charge_c

      logical          this_block(0:max_blocks), need_block(max_blocks)
      common/to_blocks/this_block              , need_block

      logical         sumgluons
      common/to_cflow/sumgluons

      logical              lwrite
      common/to_multiproc/ lwrite

      logical           jetloop
      common/to_jetloop/jetloop

      integer          matchline(0:maxlines)
      common/to_qmatch/matchline

      logical              cross_opt
      common /to_crossopt/ cross_opt

      character*70       cversion
      common/to_version/ cversion
      character*(40)       particleID, interactionID
      common/to_modelinput/particleID, interactionID

      logical             first_proc
      common/to_firstproc/first_proc


! Data Table

! If you want the output function written using real*8 and complex*16
! style declarations, remove the comment character from the following
! line.

      data fortran /77/
      data interactionid /"???"/
      data particleid /"???"/

!-----------
! Begin Code
!-----------

c kouhei lastmod Point-1 :Initial values to specify color insertions
      dipole = DIPOLESWITCH
      colorext = COLORSWITCH
      amphel = HELICITYSWITCH
c kouhei lastmod for different helicity amplitude P-1

c emit(i,j): 
c emit(1,j): emitter T or F factor
c emit(1,1): line number
c emit(1,2): species: 1=quark,2=unti-quark, 3=gluon
c emit(1,3): 1=final, 2=initial
c emit(2,j): emitter delta
c emit(2,1): line number
c emit(2,2): species: 1=quark,2=unti-quark, 3=gluon
c emit(2,3): 1=final, 2=initial
      
      emit(1,1) = EMITTERLEG
      emit(1,2) = EMITTERKIND
      emit(1,3) = EMITTERFININI

      spec(1,1) = SPECTATORLEG
      spec(1,2) = SPECTATORKIND
      spec(1,3) = SPECTATORFININI


      emit(2,1) = spec(1,1)
      emit(2,2) = spec(1,2)
      emit(2,3) = spec(1,3)
      spec(2,1) = emit(1,1)
      spec(2,2) = emit(1,2)
      spec(2,3) = emit(1,3)

      do i = 1,2
         do j = 1,3
      print*,'emit(',i,',',j,')=',emit(i,j)
         enddo
      enddo

      do i = 1,2
         do j = 1,3
      print*,'spec(',i,',',j,')=',spec(i,j)
         enddo
      enddo

c kouhei lastmod  Point-1

c      open (unit=6, file='output.dat',status='unknown')
      cversion='Generated by MadGraph II'

      first_proc = .true.
      ntopold=0
      lwrite=.true.                   !really create these files
      print*,'A resetV start -----------------------'
      call resetV
      print*,'A resetV end   ======================='
      print*,'B Load_Particle start -----------------------'
      call Load_Particles
      print*,'B Load_Particle end   ======================='
      print*,'C Load_Interactions start -----------------------'
      call Load_Interactions
      print*,'C Load_Interactions end   ======================='
      print*,' '
c      call writelogo
      print*,' '
      print*,'1 readproc start -----------------------'
 10   call readproc(more)
      print*,'1 readproc end   ======================='
      open(unit=isubfile,file='subproc.txt',status='unknown')
      write(isubfile,'($a)') 'PROCESS = '
      open(unit=22,status='scratch')
      do while (more)
         open(unit=91,status='scratch')
         flows(0,0,0,0)=0
        print*,'After readproc-1 sumgluons=',sumgluons
         sumgluons=.true.
         sumgluons=.false.
        print*,'After readpro-2 sumgluons=',sumgluons
         wcount=0
         print*,' '
         next=iline(0)
         if (lwrite) then
            open (unit=34,file='props.inc',status='unknown')
            open (unit=35,file='configs.inc',status='unknown')
            open (unit=36,file='decayBW.inc',status='unknown')
            open (unit=59,file='ngraphs.inc',status='unknown')
         else
            open (unit=34,status='scratch')
            open (unit=35,status='scratch')
            open (unit=36,status='scratch')
            open (unit=59,status='scratch')
         endif
         print*,'2 InsertFields start ----------------------- '
         call InsertFields(tops,graphs)
         print*,'2 InsertFields end    ======================='
         print*,'After InsertFields sumgluons=',sumgluons
         call write_config(tops,-99)       !Close out 35
         if (lwrite) then
            close(35)
            close(36)
            close(34)
            close(59)
         else
            close(34,status='delete')
            close(35,status='delete')
            close(36,status='delete')
            close(59,status='delete')
         endif
         if (graphs(0,0) .eq. 0) then
         else
            print*,'There are ',graphs(0,0),' graphs.'
            graphcolor(0,0,0)=graphs(0,0) !Number of graphs
            print*,' '
            call saveproc(i,graphs(0,0))
c            write(*,*) 'Not generating graphs'
c            if (abs(i) .gt. 0) then     !New Process
            if (i .gt. 0 .or. i .lt. -1) then     !New Process 
               if (lwrite) then     !(i<0  means found matched process abs(i))
                  lwrite=.false.
                  first_proc = .false.   !Now have processes
                  call write_symmetry
                  call gen_ps(graphs,tops,iline(0))

c kouhei
       print*,'graphcolor(0,0,0)=ngraph=',graphcolor(0,0,0)
       do i=1,graphcolor(0,0,0)
         print*,'graphcolor(0,0,',i,')=nterms=',graphcolor(0,0,i)
         do j=1,graphcolor(0,0,i)
         print*,'graphcolor(0',j,i,')=',graphcolor(0,j,i)
         print*,'graphcolor(1',j,i,')=',graphcolor(1,j,i)
         print*,'graphcolor(2',j,i,')=',graphcolor(2,j,i)
         enddo
      enddo

c kouhei lastmod Point-2: Selection of normal or dipole routine.
         print*,'iline(0)=',iline(0)
         if ( dipole .and. colorext) then
            print*,'3 square_mod start ----------------------- '
                  call square_mod(flows,sflows,emit,spec,iline)
              print*, 'flows(0,0,0,0)=nflows=',flows(0,0,0,0)
            print*,'3 square_mod end    ======================= '
         else
            print*,'3 square start ----------------------- '
                  call square(flows,sflows)
              print*, 'flows(0,0,0,0)=nflows=',flows(0,0,0,0)
            print*,'3 square end    ======================= '
c kouhei lastmod Point-2
         endif
c kouhei lastmod
            print*,'After square sumgluons=',sumgluons
                  open(unit=99,status='scratch')
            print*,'4 matrixcolor start ----------------------- '
                  call matrixcolor(graphcolor,sflows)
            print*,'4 matrixcolor end    ======================= '
                  rewind 99
                  rewind 91
c kouhei lastmod for different helicity amplitude P-2
         if (dipole) then
           if (colorext) then
             call writeamp_col(graphs(0,0),next,
     $                 goal_coup(2),goal_coup(1),i,flows)
           else
             if (amphel) then
               print*,'5 writeamp_mod start ----------------------- '
                  call writeamp_mod(graphs(0,0),next,
     $                 goal_coup(2),goal_coup(1),i,flows,emit)
               print*,'5 writeamp_mod end ----------------------- '
             else
               call writeamp_col(graphs(0,0),next,
     $                 goal_coup(2),goal_coup(1),i,flows)
             endif
           endif
         else
            print*,'5 writeamp start ----------------------- '
                  call writeamp(graphs(0,0),next,
     $                 goal_coup(2),goal_coup(1),i,flows)
            print*,'5 writeamp end ----------------------- '
         endif
c kouhei lastmod for different helicity 
                  close(99,status='delete')

                  do j=0,iline(0)
                     matchline(j)=iline(j)
                  enddo
                  open(unit=itnum,file='leshouche.inc',status='unknown')
                  call write_leshouche(flows, itnum, i) !(color,ilun,iproc)
                  close(itnum)
               else
                  open(unit=itnum,file='minput.mg',status='unknown',
     &                 access='append')
                  call write_proc(itnum)
                  close(itnum)
               endif
            elseif (i .eq. -1)     then !Found Match
               call genparton2()   !Write out pdf's 
c               write(*,*) 'Writing genparton2',i
               open(unit=itnum,file='leshouche.inc',status='unknown',
     &              access='append')
               call write_leshouche(flows, itnum, i) !(color,ilun,iproc)
               close(itnum)
            endif
            if (cross_opt) then   
               open(unit=itnum,file='minput.mg',status='unknown',
     &              access='append')
               call write_cross(itnum)
               close(itnum)
            endif
            close(91,status='delete')
         endif
         close(35)
c         call resetV
         write(*,*) 
         write(*,*) 
         if (jetloop .or. lwrite) then
            call readproc(more)
         else          !Only interested in processes that can match
            open(unit=itnum,file='minput.mg',status='unknown',
     &        access='append')
            lpossible = .false.
            do while (more .and. .not. lpossible) 
               call readproc(more)
               if (more) then
                  lpossible = qmatch_check(iline(0))
                  if (.not. lpossible) then
                     call write_proc(itnum)
                  endif
               endif
            enddo
            close(itnum)
         endif
               
      enddo
      rewind 22
c      call writedsig(iline(0),22)
      close(22,status='delete')
      close(unit=isubfile)
      if (.not. lwrite) then
         open(unit=itnum,file='minput.mg',status='unknown',
     &        access='append')
c         write(itnum,*)' '
         write(itnum,*) 'done'
         close(itnum)
      endif
      write(*,*)
      write(*,*)
      write(*,*) 'Thank you for using MadGraph'
      write(*,*)
      
!************* These are for color flow information only ******************
!      call writecolor(graphcolor,sflows,goodflow)
!      call writeflows(flows,goodflow)

      end



      subroutine square(flows,sflows)
!**************************************************************************
!     Square terms and simplify to get color factors for flows
!**************************************************************************
      implicit none

! Constants

      include 'params.inc'
c      integer    maxlines,  maxfactors,  maxterms,    maxflows
c      parameter (maxlines=8,maxfactors=9,maxterms=250,maxflows=200)

! Arguments

      integer flows(0:2*maxlines,0:maxfactors,0:24,0:maxflows)
      double precision  sflows(maxflows,maxflows)

! Local Variables

      integer color(0:2*maxlines,0:maxfactors,0:maxterms)
      integer nflows,nfactors,nelements
      integer iflow,ifact,ielement,iterm,cterm,fterm
      integer cflow,cfact,cfactors
      integer i,j
      double precision sdiag
c kouhei
c      integer saveflows(0:2*maxlines,0:maxfactors,0:24,0:maxflows)
      integer nterm
      double precision colorsum
      
c kouhei
      
! Global Variables

      integer           isflows(2,maxflows,maxflows)
      common/to_isflow/isflows
      logical         sumgluons
      common/to_cflow/sumgluons
!-----------
! Begin Code
! -----------
        nterm = 2
      nflows=flows(0,0,0,0)
      sumgluons=.true.
      do iflow=1,nflows
c         write(*,*) 'Flow number ',iflow
c         call PrintColors(flows(0,0,0,iflow))
         call SimpColors(flows(0,0,0,iflow),-1)  !Sum over internal gluon
c         call PrintColors(flows(0,0,0,iflow))
      enddo

      colorsum = 0.d0

      do iflow=1,nflows
         do cflow=iflow,nflows
            do iterm=1,flows(0,0,0,iflow)
               nfactors=flows(0,0,iterm,iflow)
               do ifact=0,nfactors
                  nelements=abs(flows(0,ifact,iterm,iflow))
                  do ielement=0,nelements
                     color(ielement,ifact,iterm)=
     &                    flows(ielement,ifact,iterm,iflow)
                  enddo
               enddo
            enddo                    !have all terms copied
            color(0,0,0)=flows(0,0,0,iflow)    !This is number of terms

            do cterm=2,flows(0,0,0,cflow)
               do iterm=1,flows(0,0,0,iflow)
                  call copyterm(color,iterm,-999,-999)
               enddo
            enddo              !now we have enough set for all square terms

            do cterm=1,flows(0,0,0,cflow)
               cfactors=flows(0,0,cterm,cflow)
               do iterm=1,flows(0,0,0,iflow)
                  nfactors=flows(0,0,iterm,iflow)
                  fterm = flows(0,0,0,iflow)*(cterm-1)+iterm
                  color(0,0,fterm) = nfactors+cfactors-1
                  color(0,1,fterm) = 2 !num and denom
                  color(1,1,fterm) = color(1,1,fterm)*
     &                 flows(1,1,cterm,cflow)
                  color(2,1,fterm) = color(2,1,fterm)*
     &                 flows(2,1,cterm,cflow)
                  do cfact=2,cfactors
                     nelements=abs(flows(0,cfact,cterm,cflow))
                     color(0,nfactors+cfact-1,fterm)=
     &                    flows(0,cfact,cterm,cflow)
                     if (flows(0,cfact,cterm,cflow) .lt. 0) then !T Matrix
                        color(1,cfact+nfactors-1,fterm) =
     &                       flows(2,cfact,cterm,cflow)
                        color(2,cfact+nfactors-1,fterm) =
     &                          flows(1,cfact,cterm,cflow)
                        do ielement=3,nelements
                           color(ielement,cfact+nfactors-1,fterm)=
     &                          flows(nelements-ielement+3,cfact,
     &                          cterm,cflow)
                        enddo
                     else
                        do ielement=1,nelements
                           color(ielement,cfact+nfactors-1,fterm)=
     &                          flows(nelements-ielement+1,cfact,
     &                          cterm,cflow)
                        enddo
                     endif
                  enddo
               enddo
            enddo

            call simpcolors(color,1)
            call addterms(color)

            if (color(0,0,0) .ge. 1) then
               isflows(1,iflow,cflow)=color(1,1,1)
               isflows(2,iflow,cflow)=color(2,1,1)
               isflows(1,cflow,iflow)=color(1,1,1)
               isflows(2,cflow,iflow)=color(2,1,1)
               sflows(iflow,cflow)=dble(color(1,1,1))/
     &              dble(color(2,1,1))
               sflows(cflow,iflow)=dble(color(1,1,1))/
     &              dble(color(2,1,1))

        colorsum = colorsum + sflows(iflow,cflow)
c       print*,'In times, colorsum=',colorsum
               if (color(0,0,0) .gt. 1) then
                  print*,'Error More than one term color(0,0,1)',
     &      color(0,0,1)
               elseif (color(0,0,1) .gt. 1) then !one factor
                  print*,'One term but many factors',color(0,0,1),
     &                 iflow,cflow,color(1,1,1)
                  call printcolors(color)
               endif
            else
               sflows(iflow,cflow)=0.d0
               sflows(cflow,iflow)=0.d0
            endif
c            if (iflow .eq. cflow .and. sflows(iflow,cflow) .lt. 0) then
c               write(*,'(2i4,f21.15)') iflow,cflow,sflows(iflow,cflow)
c               write(*,'(4i4)') color(1,1,1),color(2,1,1)
c            endif
         enddo
      enddo

      write(*,*) 'colorsumfinal='
      write(*,'(1x,1e16.8)') colorsum
c
c     Surprisingly diagonal elements aren't always biggest because
c     of extra terms which subtract. So don't use this check anymore
c
c      do i=1,nflows
c         sdiag = abs(sflows(i,i))
c         do j=1,nflows
c            if (abs(sflows(j,i)) .gt. sdiag) then
c               write(*,'(a,2i4,2f20.5)') 'Warning diag not largest',j,i,
c     &              sflows(j,i)/sdiag, sdiag
c               write(*,'(a,2i4,1f10.5,2i20)') 'Warning',j,i,
c     &              sflows(j,i)/sdiag, isflows(1,j,i),isflows(2,j,i)
c
c            endif
c         enddo
c      enddo
      end


      subroutine square_mod(flows,sflows,emit,spec,iline)
!**************************************************************************
!     Square terms and simplify to get color factors for flows
!**************************************************************************
      implicit none

! Constants

      include 'params.inc'
c      integer    maxlines,  maxfactors,  maxterms,    maxflows
c      parameter (maxlines=8,maxfactors=9,maxterms=250,maxflows=200)

! Arguments

      integer flows(0:2*maxlines,0:maxfactors,0:24,0:maxflows)
      double precision  sflows(maxflows,maxflows)

! Local Variables

      integer color(0:2*maxlines,0:maxfactors,0:maxterms)
      integer nflows,nfactors,nelements
      integer iflow,ifact,ielement,iterm,cterm,fterm
      integer cflow,cfact,cfactors
      integer i,j
      double precision sdiag

c kouhei lstmod
      integer flowsmod(0:2*maxlines,0:maxfactors,0:24,0:maxflows)
c      integer saveflows(0:2*maxlines,0:maxfactors,0:24,0:maxflows)
      integer nterm
      integer ntermcolmod,nfactorscolmod(99)
      integer emit(2,3)
      integer spec(2,3)
      integer iline(-maxlines:maxlines)
      double precision colorsum
c kouhei
      
! Global Variables

      integer           isflows(2,maxflows,maxflows)
      common/to_isflow/isflows
      logical         sumgluons
      common/to_cflow/sumgluons

c kouhei lastmod
      common/to_fmod/flowsmod
c kouhei lastmod

!-----------
! Begin Code
! -----------
 
        nterm = 2
      nflows=flows(0,0,0,0)
c kouhei
c         print*,'Original information of term number'
c         print*,'Total flow number=nflows',nflows
c        do iflow=1,nflows
c         print*, 'Term number of flow=',iflow,':',flows(0,0,0,iflow)
c        enddo
c               print*,'Original information'
c               print*,'nflows=',nflows 
c               do cflow=1,nflows
c             print*,'cflow=',cflow 
c           call flowscomponents(nterm,1,cflow,flows)
c               enddo  
c kouhei end
      sumgluons=.true.
      do iflow=1,nflows
c         write(*,*) 'Flow number ',iflow
c         call PrintColors(flows(0,0,0,iflow))
c      print*,'3-1 SimpColors(flows,-1) start ----------------- '
         call SimpColors(flows(0,0,0,iflow),-1)  !Sum over internal gluon
c      print*,'3-1 SimpColors(flows,-1) end =================== '
c         call PrintColors(flows(0,0,0,iflow))
      enddo
c kouhei
c               print*,'After SimpColors(flows,-1)'
c               print*,'nflows=',nflows 
               do cflow=1,nflows
c           print*,'cflow=',cflow 
           call flowscomponents(nterm,1,cflow,flows)
               enddo  
c kouhei

c kouhei lastmod: Copy flows to flowsmod
      do iflow=1,nflows
            do iterm=1,flows(0,0,0,iflow)
               nfactors=flows(0,0,iterm,iflow)
               do ifact=0,nfactors
                  nelements=abs(flows(0,ifact,iterm,iflow))
                  do ielement=0,nelements
                     flowsmod(ielement,ifact,iterm,iflow)=
     &                    flows(ielement,ifact,iterm,iflow)
                  enddo
               enddo
            enddo                    !have all terms copied
         flowsmod(0,0,0,iflow)=flows(0,0,0,iflow)    !This is number of terms
c       print*,'After copy from flows to flowsmod,iflow=',iflow
c       call flowsmodcomponents(2,1,iflow,flowsmod)

C kouhei Point-3: Color insertion to spectator
          Print*,'Spectator insertion'
          Print*,'Before T or F delta insertion to spectater:flows'
          if    (  (spec(2,2) .eq. 1 .and. spec(2,3) .eq. 1) .or.
     &             (spec(2,2) .eq. 2 .and. spec(2,3) .eq. 2)) then
          Print*,'T delta insertion to spectater : flows'
                 call addTdel(flowsmod(0,0,0,iflow),70,spec(2,1))
          elseif ( (spec(2,2) .eq. 1 .and. spec(2,3) .eq. 2) .or.
     &             (spec(2,2) .eq. 2 .and. spec(2,3) .eq. 1)) then
          Print*,'T delta insertion to spectater : flows'
                 call addTdel(flowsmod(0,0,0,iflow),spec(2,1),70)
          elseif ( spec(2,2) .eq. 3 ) then
          Print*,'F delta insertion to spectater : flows'
                 call addFdel(flowsmod(0,0,0,iflow),70,spec(2,1))
          else 
              print*,'Error: unknown spectater delta'
          endif
          Print*,'After T or F delta insertion to spectater:flows'
c          call printcolors(flowsmod(0,0,0,iflow))
          call SumColors(flowsmod(0,0,0,iflow),spec(2,1))

          Print*,'Before T or F factor insertion to spectater:flows'
          if    (  (spec(1,2) .eq. 1 .and. spec(1,3) .eq. 1) .or.
     &             (spec(1,2) .eq. 2 .and. spec(1,3) .eq. 2)) then
          Print*,'T factor insertion to spectater : flows'
                 call addTfactor(flowsmod(0,0,0,iflow),80,90,spec(1,1))
          elseif ( (spec(1,2) .eq. 1 .and. spec(1,3) .eq. 2) .or.
     &             (spec(1,2) .eq. 2 .and. spec(1,3) .eq. 1)) then
         Print*,'T factor insertion to spectater : flows'
                 call addTfactor(flowsmod(0,0,0,iflow),80,spec(1,1),90)
          elseif ( spec(1,2) .eq. 3 .and. spec(1,3) .eq. 1) then
         Print*,'F factor insertion to spectater : flows'
                  call addFfactor(flowsmod(0,0,0,iflow),80,90,spec(1,1))
          elseif ( spec(1,2) .eq. 3 .and. spec(1,3) .eq. 2) then
         Print*,'F factor insertion to spectater : flows'
                  call addFfactor(flowsmod(0,0,0,iflow),80,90,spec(1,1))
          else 
              print*,'Error: unknown spectater T or F factor'
          endif
          Print*,'After Ffactor insertion to spectater:flows'
c          call printcolors(flowsmod(0,0,0,iflow))
          call SumColors(flowsmod(0,0,0,iflow),spec(1,1))

C kouhei Point-3 end 

c       print*,'After color insertion to flowsmod'
c      call printcolors(flowsmod(0,0,0,iflow))
c       call flowsmodcomponents(2,1,iflow,flowsmod)
      enddo
c      print*,'End of insertion to flowsmod'
c      do iflow=1,nflows
c      print*,'iflow=',iflow
c      call printcolors(flowsmod(0,0,0,iflow))
c      enddo

c kouhei lastmod

      colorsum = 0.d0
      do iflow=1,nflows
         do cflow=iflow,nflows

      print*,' ======================================== '
      print*,' New CF(iflow,cflow) ==================== '
      print*,'(iflow,cflow)=(',iflow,',',cflow,')'

            do iterm=1,flows(0,0,0,iflow)
               nfactors=flows(0,0,iterm,iflow)
               do ifact=0,nfactors
                  nelements=abs(flows(0,ifact,iterm,iflow))
                  do ielement=0,nelements
                     color(ielement,ifact,iterm)=
     &                    flows(ielement,ifact,iterm,iflow)
                  enddo
               enddo
            enddo                    !have all terms copied
            color(0,0,0)=flows(0,0,0,iflow)    !This is number of terms

c kouhei
c               print*,'After copy from flows to color'
c               call printcolors(color)
c kouhei

C kouhei Point-4: Color insertion to emitter
          Print*,'Emitter insertion'
          Print*,'Before T or F delta insertion to emitter:color'
          if    (  (emit(2,2) .eq. 1 .and. emit(2,3) .eq. 1) .or.
     &             (emit(2,2) .eq. 2 .and. emit(2,3) .eq. 2)) then
          Print*,'T delta insertion to emitter : color'
                 call addTdel(color,90,emit(2,1))
          elseif ( (emit(2,2) .eq. 1 .and. emit(2,3) .eq. 2) .or.
     &             (emit(2,2) .eq. 2 .and. emit(2,3) .eq. 1)) then
          Print*,'T delta insertion to emitter : color'
                 call addTdel(color,emit(2,1),90)
          elseif ( emit(2,2) .eq. 3 ) then
          Print*,'F delta insertion to emitter : color'
                 call addFdel(color,90,emit(2,1))
          else 
              print*,'Error: unknown emitter delta'
          endif
          Print*,'After T or F delta insertion to emitter:color'
c          call printcolors(color)
          call SumColors(color,emit(2,1))

          Print*,'Before T or F factor insertion to emitter:color'
          if    (  (emit(1,2) .eq. 1 .and. emit(1,3) .eq. 1) .or.
     &             (emit(1,2) .eq. 2 .and. emit(1,3) .eq. 2)) then
          Print*,'T factor insertion to emitter : color'
                 call addTfactor(color,80,70,emit(1,1))
          elseif ( (emit(1,2) .eq. 1 .and. emit(1,3) .eq. 2) .or.
     &             (emit(1,2) .eq. 2 .and. emit(1,3) .eq. 1)) then
         Print*,'T factor insertion to emitter : color'
                 call addTfactor(color,80,emit(1,1),70)
          elseif ( emit(1,2) .eq. 3 .and. emit(1,3) .eq. 1) then
         Print*,'F factor insertion to emitter : color'
                call addFfactor(color,70,emit(1,1),80)
          elseif ( emit(1,2) .eq. 3 .and. emit(1,3) .eq. 2) then
         Print*,'F factor insertion to emitter : color'
                call addFfactor(color,70,emit(1,1),80)
          else 
              print*,'Error: unknown emitter T or F factor'
          endif
          Print*,'After Ffactor insertion to emitter:color'
c          call printcolors(color)
          call SumColors(color,emit(1,1))
C kouhei Point-4 end

c           print*,'After emitter insertion to color'
c               print*,'iflow=',iflow
c               call printcolors(color)
c               call colorcomponents(nterm,iflow,cflow,color)

c           print*,'After spectator insertion to flowsmod(0,0,0,cflow)'
c              print*,'cflow=',cflow
c               call printcolors(flowsmod(0,0,0,cflow))
c               call flowsmodcomponents(2,1,cflow,flowsmod)

c Terms and factors of color(i,j,k)
           ntermcolmod=color(0,0,0)
c          print*,'ntermcolmod=',ntermcolmod
           do i=1,ntermcolmod
             nfactorscolmod(i)=color(0,0,i)
c          print*,'nfactorscolmod(',i,')=',nfactorscolmod(i)
           enddo
c kouhei lastmod


c kouhei lastmod : 
c            do cterm=2,flows(0,0,0,cflow)
            do cterm=2,flowsmod(0,0,0,cflow)
c               do iterm=1,flows(0,0,0,iflow)
               do iterm=1,ntermcolmod
c kouhei lastmod : 
c      print*,'3-2 copyterm start ----------------- '
                  call copyterm(color,iterm,-999,-999)
c      print*,'3-2 copyterm end =================== '
               enddo
            enddo              !now we have enough set for all square terms
c                print*,'After copyterm'
c               call printcolors(color)
c                 call colorcomponents(nterm,iflow,cflow,color) 

c kouhei lastmod:
c            do cterm=1,flows(0,0,0,cflow)
           do cterm=1,flowsmod(0,0,0,cflow)
c               cfactors=flows(0,0,cterm,cflow)
               cfactors=flowsmod(0,0,cterm,cflow)
c               do iterm=1,flows(0,0,0,iflow)
               do iterm=1,ntermcolmod
c                  nfactors=flows(0,0,iterm,iflow)
                  nfactors=nfactorscolmod(iterm)
c                  fterm = flows(0,0,0,iflow)*(cterm-1)+iterm
                  fterm = ntermcolmod*(cterm-1)+iterm
                  color(0,0,fterm) = nfactors+cfactors-1
                  color(0,1,fterm) = 2 !num and denom
c                  color(1,1,fterm) = color(1,1,fterm)*
c     &                 flows(1,1,cterm,cflow)
                  color(1,1,fterm) = color(1,1,fterm)*
     &                 flowsmod(1,1,cterm,cflow)
c                  color(2,1,fterm) = color(2,1,fterm)*
c     &                 flows(2,1,cterm,cflow)
                  color(2,1,fterm) = color(2,1,fterm)*
     &                 flowsmod(2,1,cterm,cflow)
                  do cfact=2,cfactors
c                     nelements=abs(flows(0,cfact,cterm,cflow))
                     nelements=abs(flowsmod(0,cfact,cterm,cflow))
c                     color(0,nfactors+cfact-1,fterm)=
c     &                    flows(0,cfact,cterm,cflow)
                     color(0,nfactors+cfact-1,fterm)=
     &                    flowsmod(0,cfact,cterm,cflow)
c                     if (flows(0,cfact,cterm,cflow) .lt. 0) then !T Matrix
c                        color(1,cfact+nfactors-1,fterm) =
c     &                       flows(2,cfact,cterm,cflow)
c                        color(2,cfact+nfactors-1,fterm) =
c     &                          flows(1,cfact,cterm,cflow)
               if (flowsmod(0,cfact,cterm,cflow) .lt. 0) then !T Matrix
                        color(1,cfact+nfactors-1,fterm) =
     &                       flowsmod(2,cfact,cterm,cflow)
                        color(2,cfact+nfactors-1,fterm) =
     &                          flowsmod(1,cfact,cterm,cflow)

                       do ielement=3,nelements
                           color(ielement,cfact+nfactors-1,fterm)=
c     &                          flows(nelements-ielement+3,cfact,
     &                          flowsmod(nelements-ielement+3,cfact,
     &                          cterm,cflow)
                        enddo
                     else
                        do ielement=1,nelements
                           color(ielement,cfact+nfactors-1,fterm)=
c     &                          flows(nelements-ielement+1,cfact,
     &                          flowsmod(nelements-ielement+1,cfact,
     &                          cterm,cflow)
                        enddo
                     endif
                  enddo
               enddo
            enddo
c kouhei lastmod:

c kouhei
c                  print*,'Multiply, conjugate and before external sum'
c             print*,'(iflow,cflow)=(',iflow,',',cflow,')'
c               call printcolors(color)
c               call colorcomponents(nterm,iflow,cflow,color) 
c kouhei

c       print*,'3-3mod External SumColors start ------------ '
c       print*,'(iflow,cflow)=(',iflow,',',cflow,')'

c kouhei Point-5 : External line color sums
c       print*,'iline(0)=',iline(0)
c       print*,'emit(1,1),spec(1,1)',emit(1,1),spec(1,1)
       do i = 1,iline(0)
          if (i .ne. emit(1,1) .and. i .ne. spec(1,1)) then
c       print*,'Sum of extenal line i=',i
            call SumColors(color,i)
c       print*,'orderF start --------------------- '
c       call orderF(color)
c       print*,'orderF end ======================= '
          endif
c       print*,'I am after sum of original external line=',i
c       print*,'in (iflow,cflow)=(',iflow,',',cflow,')'

       enddo

c       print*,'Sum of extenal line 70'
            call SumColors(color,70)
c       print*,'I am after sum of added external line 70'
c       print*,'in (iflow,cflow)=(',iflow,',',cflow,')'
c       print*,'Sum of extenal line 90'
            call SumColors(color,90)
c       print*,'I am after sum of added external line 90'
c       print*,'in (iflow,cflow)=(',iflow,',',cflow,')'
c       print*,'Sum of extenal line 80'
            call SumColors(color,80)
c kouhei Point-5 end

c       print*,'3-3mod External SumColors end ------------ '
c                  print*,'After external sum'
c       print*,'(iflow,cflow)=(',iflow,',',cflow,')'
c               call printcolors(color)
c                  call colorcomponents(nterm,iflow,cflow,color) 
c kouhei

C Kouhei
c      print*,'3-4 addterms start ----------------- '
            call addterms(color)
c      print*,'3-4 addterms end =================== '

c                print*,'After addterms'
c               call printcolors(color)
c                 call colorcomponents(nterm,iflow,cflow,color) 
c kouhei
            if (color(0,0,0) .ge. 1) then
              if ( (spec(1,2) .eq. 1 .and. spec(1,3) .eq. 2) .or.
     &            (spec(1,2) .eq. 2 .and. spec(1,3) .eq. 1)) then
                  color(1,1,1)=-color(1,1,1)
              endif
              if ( (emit(1,2) .eq. 1 .and. emit(1,3) .eq. 2) .or.
     &             (emit(1,2) .eq. 2 .and. emit(1,3) .eq. 1)) then
                  color(1,1,1)=-color(1,1,1)
              endif
               isflows(1,iflow,cflow)=color(1,1,1)
               isflows(2,iflow,cflow)=color(2,1,1)
               isflows(1,cflow,iflow)=color(1,1,1)
               isflows(2,cflow,iflow)=color(2,1,1)
               sflows(iflow,cflow)=dble(color(1,1,1))/
     &              dble(color(2,1,1))
               sflows(cflow,iflow)=dble(color(1,1,1))/
     &              dble(color(2,1,1))
c kouhei
        colorsum = colorsum + sflows(iflow,cflow)
c        print*,'In times, colorsum=',colorsum
c                  print*,'At end of square'
c               call printcolors(color)
c                  call colorcomponents(nterm,iflow,cflow,color) 
c       print*,'isflows({1,2},iflow,cflow)'
c       print*,'isflows(1',iflow,',',cflow,')=',isflows(1,iflow,cflow)
c       print*,'isflows(2',iflow,',',cflow,')=',isflows(2,iflow,cflow)
c kouhei
               if (color(0,0,0) .gt. 1) then
                  print*,'Error More than one term color(0,0,1)',
     &      color(0,0,1)
               elseif (color(0,0,1) .gt. 1) then !one factor
                  print*,'One term but many factors',color(0,0,1),
     &                 iflow,cflow,color(1,1,1)
                  call printcolors(color)
               endif
            else
               sflows(iflow,cflow)=0.d0
               sflows(cflow,iflow)=0.d0
            endif
c            if (iflow .eq. cflow .and. sflows(iflow,cflow) .lt. 0) then
c               write(*,'(2i4,f21.15)') iflow,cflow,sflows(iflow,cflow)
c               write(*,'(4i4)') color(1,1,1),color(2,1,1)
c            endif
         enddo
      enddo

c      write(*,*) 'colorsumfinal='
c      write(*,'(1x,1e16.8)') colorsum
c
c     Surprisingly diagonal elements aren't always biggest because
c     of extra terms which subtract. So don't use this check anymore
c
c      do i=1,nflows
c         sdiag = abs(sflows(i,i))
c         do j=1,nflows
c            if (abs(sflows(j,i)) .gt. sdiag) then
c               write(*,'(a,2i4,2f20.5)') 'Warning diag not largest',j,i,
c     &              sflows(j,i)/sdiag, sdiag
c               write(*,'(a,2i4,1f10.5,2i20)') 'Warning',j,i,
c     &              sflows(j,i)/sdiag, isflows(1,j,i),isflows(2,j,i)
c
c            endif
c         enddo
c      enddo
      end

               


      subroutine flowscomponents(dnterm,diflow,dcflow,dflows)
      implicit none

      include 'params.inc'

      integer i,j,diflow,dcflow,dnterm
      integer dflows(0:2*maxlines,0:maxfactors,0:24,0:maxflows)

c      print*,'(iflow,cflow)=',diflow,dcflow
      print*,'Terms=flows(0,0,0,',dcflow,')',dflows(0,0,0,dcflow)
      do i = 1,dnterm
           print*,'flows(0,0,',i,',',dcflow,')',dflows(0,0,i,dcflow)
      enddo
      do i = 1,dnterm
            print*,'term=',i
            print*,'flows(1,1,',i,',',dcflow,')',dflows(1,1,i,dcflow)
            print*,'flows(2,1,',i,',',dcflow,')',dflows(2,1,i,dcflow)
            print*,'flows(3,1,',i,',',dcflow,')',dflows(3,1,i,dcflow)
            print*,'flows(4,1,',i,',',dcflow,')',dflows(4,1,i,dcflow)
            print*,'flows(0,1,',i,',',dcflow,')',dflows(0,1,i,dcflow)
              do j=2,4
                  print*,'Vertex kind of',j,'-th factor in ',
     &  i,'-th term at',dcflow,'-th cflow'
        print*,'flows(0,',j,',',i,',',dcflow,')',dflows(0,j,i,dcflow)
              enddo
              do j=2,4
                print*,'Fields of',j,'-th factor in',i,'-th term at',
     &  dcflow,'-th cflow'
         print*,'flows(1,',j,',',i,',',dcflow,')',dflows(1,j,i,dcflow)
         print*,'flows(2,',j,',',i,',',dcflow,')',dflows(2,j,i,dcflow)
         print*,'flows(3,',j,',',i,',',dcflow,')',dflows(3,j,i,dcflow)
         print*,'flows(4,',j,',',i,',',dcflow,')',dflows(4,j,i,dcflow)
         print*,'flows(5,',j,',',i,',',dcflow,')',dflows(5,j,i,dcflow)
              enddo
      enddo
        call printcolors(dflows(0,0,0,dcflow))
      end

      subroutine colorcomponents(dnterm,diflow,dcflow,dcolor)
      implicit none

      include 'params.inc'

      integer i,j,diflow,dcflow,dnterm
      integer dcolor(0:2*maxlines,0:maxfactors,0:maxterms)

                  print*,'(iflow,cflow)=',diflow,dcflow
      print*,'Terms=color(0,0,0)',dcolor(0,0,0)
                 do i = 1,dnterm
                   print*,'color(0,0,',i,')',dcolor(0,0,i)
                 enddo

                 do i = 1,dnterm
             print*,'term=',i
                   print*,'color(1,1,',i,')',dcolor(1,1,i)
                   print*,'color(2,1,',i,')',dcolor(2,1,i)
                   print*,'color(3,1,',i,')',dcolor(3,1,i)
                   print*,'color(4,1,',i,')',dcolor(4,1,i)
                   print*,'color(0,1,',i,')',dcolor(0,1,i)
                     do j=2,4
                   print*,'Kind of vertex of',j,'-th factor in ',
     &  i,'-th term'
                   print*,'color(0,',j,',',i,')',dcolor(0,j,i)
                     enddo
                     do j=2,4
             print*,'Fields of',j,'-th factor in',i,'-th term'
             print*,'color(1,',j,',',i,')',dcolor(1,j,i)
             print*,'color(2,',j,',',i,')',dcolor(2,j,i)
             print*,'color(3,',j,',',i,')',dcolor(3,j,i)
             print*,'color(4,',j,',',i,')',dcolor(4,j,i)
             print*,'color(5,',j,',',i,')',dcolor(5,j,i)
                     enddo
                 enddo
                  call printcolors(dcolor(0,0,0))
      end


      subroutine flowsmodcomponents(dnterm,diflow,dcflow,dflows)
      implicit none

      include 'params.inc'

      integer i,j,diflow,dcflow,dnterm
      integer dflows(0:2*maxlines,0:maxfactors,0:24,0:maxflows)

c      print*,'(iflow,cflow)=',diflow,dcflow
      print*,'Terms=flowsmod(0,0,0,',dcflow,')',dflows(0,0,0,dcflow)
      do i = 1,dnterm
           print*,'flowsmod(0,0,',i,',',dcflow,')',dflows(0,0,i,dcflow)
      enddo
      do i = 1,dnterm
           print*,'term=',i
           print*,'flowsmod(1,1,',i,',',dcflow,')',dflows(1,1,i,dcflow)
           print*,'flowsmod(2,1,',i,',',dcflow,')',dflows(2,1,i,dcflow)
           print*,'flowsmod(3,1,',i,',',dcflow,')',dflows(3,1,i,dcflow)
           print*,'flowsmod(4,1,',i,',',dcflow,')',dflows(4,1,i,dcflow)
           print*,'flowsmod(0,1,',i,',',dcflow,')',dflows(0,1,i,dcflow)
              do j=2,4
                  print*,'Vertex kind of',j,'-th factor in ',
     &  i,'-th term at',dcflow,'-th cflow'
        print*,'flows(0,',j,',',i,',',dcflow,')',dflows(0,j,i,dcflow)
              enddo
              do j=2,4
                print*,'Fields of',j,'-th factor in',i,'-th term at',
     &  dcflow,'-th cflow'
       print*,'flowsmod(1,',j,',',i,',',dcflow,')',dflows(1,j,i,dcflow)
       print*,'flowsmod(2,',j,',',i,',',dcflow,')',dflows(2,j,i,dcflow)
       print*,'flowsmod(3,',j,',',i,',',dcflow,')',dflows(3,j,i,dcflow)
       print*,'flowsmod(4,',j,',',i,',',dcflow,')',dflows(4,j,i,dcflow)
       print*,'flowsmod(5,',j,',',i,',',dcflow,')',dflows(5,j,i,dcflow)
              enddo
      enddo
        call printcolors(dflows(0,0,0,dcflow))
      end


